"""
A docker executor for Jaypore CI.
"""
import json
import subprocess

import pendulum
from rich import print as rprint

from jaypore_ci.interfaces import Executor, TriggerFailed, JobStatus
from jaypore_ci.logging import logger


def __check_output__(cmd):
    """
    Common arguments that need to be provided while
    calling subprocess.check_output
    """
    proc = subprocess.run(
        cmd, check=False, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.PIPE
    )
    if proc.returncode != 0:
        print(proc.stdout.decode())
        raise TriggerFailed(cmd)
    return proc.stdout.decode().strip()


class Docker(Executor):
    """
    Run jobs via docker.

    This will:
        - Create a separate network for each run
        - Run jobs as part of the network
        - Clean up all jobs when the pipeline exits.
    """

    def __init__(self):
        super().__init__()
        self.pipe_id = None
        self.pipeline = None

    def logging(self):
        """
        Returns a logging instance that has executor specific
        information bound to it.
        """
        return logger.bind(pipe_id=self.pipe_id, network_name=self.get_net())

    def set_pipeline(self, pipeline):
        """
        Set executor's pipeline to the given one.

        This will clean up old networks and create new ones.
        """
        if self.pipe_id is not None:
            self.delete_network()
            self.delete_all_jobs()
        self.pipe_id = pipeline.pipe_id
        self.pipeline = pipeline
        self.create_network()

    def __exit__(self, exc_type, exc_value, traceback):
        self.delete_network()
        self.delete_all_jobs()

    def get_net(self):
        """
        Return a network name based on what the curent pipeline is.
        """
        return f"jayporeci__net__{self.pipe_id}" if self.pipe_id is not None else None

    def create_network(self):
        """
        Will create a docker network.

        If it fails to do so in 3 attempts it will raise an
        exception and fail.
        """
        assert self.pipe_id is not None, "Cannot create network if pipe is not set"
        for _ in range(3):
            net_ls = subprocess.run(
                f"docker network ls | grep {self.get_net()}",
                shell=True,
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if net_ls.returncode == 0:
                self.logging().info(
                    "Found network", network_name=self.get_net(), subprocess=net_ls
                )
                return net_ls
            self.logging().info(
                "Create network",
                subprocess=__check_output__(
                    f"docker network create -d bridge {self.get_net()}"
                ),
            )
        raise TriggerFailed("Cannot create network")

    def delete_all_jobs(self):
        """
        Deletes all jobs associated with the pipeline for this
        executor.

        It will stop any jobs that are still running.
        """
        assert self.pipe_id is not None, "Cannot delete jobs if pipe is not set"
        job = None
        for job in self.pipeline.jobs.values():
            if job.run_id is not None and not job.run_id.startswith("pyrun_"):
                self.logging().info(
                    "Stop job:",
                    subprocess=__check_output__(f"docker stop -t 1 {job.run_id}"),
                )
                job.check_job(with_update_report=False)
        if job is not None:
            job.check_job()
        self.logging().info("All jobs stopped")

    def delete_network(self):
        """
        Delete the network for this executor.
        """
        assert self.pipe_id is not None, "Cannot delete network if pipe is not set"
        self.logging().info(
            "Delete network",
            subprocess=__check_output__(
                f"docker network rm {self.get_net()} || echo 'No such net'"
            ),
        )

    def get_job_name(self, job):
        """
        Generates a clean job name slug.
        """
        name = "".join(
            l if l in "abcdefghijklmnopqrstuvwxyz1234567890" else "-"
            for l in job.name.lower()
        )
        return f"jayporeci__job__{self.pipe_id}__{name}"

    def run(self, job: "Job") -> str:
        """
        Run the given job and return a docker container ID.
        In case something goes wrong it will raise TriggerFailed
        """
        assert self.pipe_id is not None, "Cannot run job if pipe id is not set"
        env_vars = [f"--env {key}={val}" for key, val in job.get_env().items()]
        trigger = [
            "docker run -d",
            "-v /var/run/docker.sock:/var/run/docker.sock",
            "-v /usr/bin/docker:/usr/bin/docker:ro",
            f"-v /tmp/jayporeci__src__{self.pipeline.remote.sha}:/jaypore_ci/run",
            *["--workdir /jaypore_ci/run" if not job.is_service else None],
            f"--name {self.get_job_name(job)}",
            f"--network {self.get_net()}",
            *env_vars,
            job.image,
            job.command if not job.is_service else None,
        ]
        if not job.is_service:
            assert job.command
        rprint(trigger)
        trigger = " ".join(t for t in trigger if t is not None)
        run_job = subprocess.run(
            trigger,
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        if run_job.returncode == 0:
            return run_job.stdout.decode().strip()
        raise TriggerFailed(run_job)

    def get_status(self, run_id: str) -> JobStatus:
        """
        Given a run_id, it will get the status for that run.
        """
        inspect = json.loads(__check_output__(f"docker inspect {run_id}"))[0]
        status = JobStatus(
            is_running=inspect["State"]["Running"],
            exit_code=int(inspect["State"]["ExitCode"]),
            logs="",
            started_at=pendulum.parse(inspect["State"]["StartedAt"]),
            finished_at=pendulum.parse(inspect["State"]["FinishedAt"])
            if inspect["State"]["FinishedAt"] != "0001-01-01T00:00:00Z"
            else None,
        )
        # --- logs
        self.logging().debug("Check status", status=status)
        logs = __check_output__(f"docker logs {run_id}")
        return status._replace(logs=logs)



"""
The code submodule for Jaypore CI.
"""
import time
import os
import subprocess
from itertools import product
from collections import defaultdict
from typing import List, Union, Callable
from contextlib import contextmanager

import structlog
import pendulum

from jaypore_ci import remotes, executors, reporters, repos
from jaypore_ci.interfaces import (
    Remote,
    Executor,
    Reporter,
    TriggerFailed,
    Status,
    Repo,
)
from jaypore_ci.logging import logger

TZ = "UTC"

__all__ = ["Pipeline", "Job"]


# All of these statuses are considered "finished" statuses
FIN_STATUSES = (Status.FAILED, Status.PASSED, Status.TIMEOUT, Status.SKIPPED)
PREFIX = "JAYPORE_"


class Job:  # pylint: disable=too-many-instance-attributes
    """
    This is the fundamental building block for running jobs.
    Each job goes through a lifecycle defined by
    :class:`~jaypore_ci.interfaces.Status`.

    A job is run by an :class:`~jaypore_ci.interfaces.Executor` as part of a
    :class:`~jaypore_ci.jci.Pipeline`.

    It is never created manually. The correct way to create a job is to use
    :meth:`~jaypore_ci.jci.Pipeline.job`.

    :param name: The name for the job. Names must be unique across jobs and stages.
    :param command: The command that we need to run for the job. It can be set
        to `None` when `is_service` is True.
    :param is_service: Is this job a service or not? Service jobs are assumed
        to be :class:`~jaypore_ci.interfaces.Status.PASSED` as long as they start.
        They are shut down when the entire pipeline has finished executing.
    :param pipeline: The pipeline this job is associated with.
    :param status: The :class:`~jaypore_ci.interfaces.Status` of this job.
    :param image: What docker image to use for this job.
    :param timeout: Defines how long a job is allowed to run before being
        killed and marked as class:`~jaypore_ci.interfaces.Status.FAILED`.
    :param env: A dictionary of environment variables to pass to the docker run command.
    :param children: Defines which jobs depend on this job's output status.
    :param parents: Defines which jobs need to pass before this job can be run.
    :param stage: What stage the job belongs to. This stage name must exist so
        that we can assign jobs to it.
    """

    def __init__(
        self,
        name: str,
        command: Union[str, Callable],
        pipeline: "Pipeline",
        *,
        status: str = None,
        image: str = None,
        timeout: int = None,
        env: dict = None,
        children: List["Job"] = None,
        parents: List["Job"] = None,
        is_service: bool = False,
        stage: str = None,
    ):
        self.name = name
        self.command = command
        self.image = image
        self.status = status
        self.run_state = None
        self.timeout = timeout
        self.pipeline = pipeline
        self.env = env
        self.children = children if children is not None else []
        self.parents = parents if parents is not None else []
        self.is_service = is_service
        self.stage = stage
        # --- run information
        self.logs = defaultdict(list)
        self.job_id = id(self)
        self.run_id = None
        self.run_start = None
        self.last_check = None

    def logging(self):
        """
        Returns a logging instance that has job specific information bound to
        it.
        """
        return self.pipeline.logging().bind(
            job_id=self.job_id,
            job_name=self.name,
            run_id=self.run_id,
        )

    def update_report(self) -> str:
        """
        Update the status report. Usually called when a job changes some of
        it's internal state like when logs are updated or when status has
        changed.
        """
        self.logging().debug("Update report")
        status = {
            Status.PENDING: "pending",
            Status.RUNNING: "pending",
            Status.FAILED: "failure",
            Status.PASSED: "success",
            Status.TIMEOUT: "warning",
            Status.SKIPPED: "warning",
        }[self.pipeline.get_status()]
        report = self.pipeline.reporter.render(self.pipeline)
        with open("/jaypore_ci/run/jaypore_ci.status.txt", "w", encoding="utf-8") as fl:
            fl.write(report)
        try:
            self.pipeline.remote.publish(report, status)
        except Exception as e:  # pylint: disable=broad-except
            self.logging().exception(e)
        return report

    def trigger(self):
        """
        Trigger the job via the pipeline's executor.
        This will immediately return and will not wait for the job to finish.

        It is also idempotent. Calling this multiple times will only trigger
        the job once.
        """
        if self.status == Status.PENDING:
            self.run_start = pendulum.now(TZ)
            self.logging().info("Trigger called")
            self.status = Status.RUNNING
            if isinstance(self.command, str):
                try:
                    self.run_id = self.pipeline.executor.run(self)
                    self.logging().info("Trigger done")
                except TriggerFailed as e:
                    job_run = e.args[0]
                    self.logging().error(
                        "Trigger failed",
                        returncode=job_run.returncode,
                        job_name=self.name,
                    )
                    logs = job_run.stdout.decode()
                    self.logs["stdout"] = reporters.clean_logs(logs)
                    self.status = Status.FAILED
        else:
            self.logging().info("Trigger called but job already running")
        self.check_job()

    def check_job(self, *, with_update_report=True):
        """
        This will check the status of the job.
        If `with_update_report` is False, it will not push an update to the remote.
        """
        if isinstance(self.command, str) and self.run_id is not None:
            self.logging().debug("Checking job run")
            self.run_state = self.pipeline.executor.get_status(self.run_id)
            self.last_check = pendulum.now(TZ)
            self.logging().debug(
                "Job run status found",
                is_running=self.run_state.is_running,
                exit_code=self.run_state.exit_code,
            )
            if self.run_state.is_running:
                self.status = Status.RUNNING if not self.is_service else Status.PASSED
            else:
                self.status = (
                    Status.PASSED if self.run_state.exit_code == 0 else Status.FAILED
                )
            self.logs["stdout"] = reporters.clean_logs(self.run_state.logs)
            if with_update_report:
                self.update_report()

    def is_complete(self) -> bool:
        """
        Is this job complete? It could have passed/ failed etc.
        We no longer need to check for updates in a complete job.
        """
        return self.status in FIN_STATUSES

    def get_env(self):
        """
        Gets the environment variables for a given job.
        Order of precedence for setting values is:

        1. Pipeline
        2. Stage
        3. Job
        """
        return {
            **{
                k[len(PREFIX) :]: v
                for k, v in os.environ.items()
                if k.startswith(PREFIX)
            },
            **self.pipeline.pipe_kwargs.get("env", {}),
            **self.env,
        }


class Pipeline:  # pylint: disable=too-many-instance-attributes
    """
    A pipeline acts as a controlling/organizing mechanism for multiple jobs.

    :param repo         : Provides information about the codebase.
    :param reporter     : Provides reports based on the state of the pipeline.
    :param remote       : Allows us to publish reports to somewhere like gitea/email.
    :param executor     : Runs the specified jobs.
    :param poll_interval: Defines how frequently (in seconds) to check the
        pipeline status and publish a report.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        repo: Repo = None,
        remote: Remote = None,
        executor: Executor = None,
        reporter: Reporter = None,
        poll_interval: int = 10,
        **kwargs,
    ) -> "Pipeline":
        self.jobs = {}
        self.services = []
        self.should_pass_called = set()
        self.repo = repo if repo is not None else repos.Git.from_env()
        self.remote = (
            remote
            if remote is not None
            else remotes.gitea.Gitea.from_env(repo=self.repo)
        )
        self.executor = executor if executor is not None else executors.docker.Docker()
        self.reporter = reporter if reporter is not None else reporters.text.Text()
        self.poll_interval = poll_interval
        self.stages = ["Pipeline"]
        self.pipe_id = (
            subprocess.check_output(
                "cat /proc/self/cgroup | grep name= | awk -F/ '{print $3}'",
                shell=True,
                stderr=subprocess.STDOUT,
            )
            .decode()
            .strip()
        )
        self.executor.set_pipeline(self)
        # ---
        kwargs["image"] = kwargs.get("image", "arjoonn/jci:latest")
        kwargs["timeout"] = kwargs.get("timeout", 15 * 60)
        kwargs["env"] = kwargs.get("env", {})
        kwargs["stage"] = "Pipeline"
        self.pipe_kwargs = kwargs
        self.stage_kwargs = None

    def logging(self):
        """
        Return a logger with information about the current pipeline bound to
        it.
        """
        return logger.bind(
            **{
                **structlog.get_context(self.remote.logging()),
                **structlog.get_context(self.executor.logging()),
                "pipe_id": id(self),
            }
        )

    def __enter__(self):
        self.executor.__enter__()
        self.remote.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.run()
        self.executor.__exit__(exc_type, exc_value, traceback)
        self.remote.__exit__(exc_type, exc_value, traceback)
        return False

    def get_status(self) -> Status:
        """
        Calculates a pipeline's status based on the status of it's jobs.
        """
        for job in self.jobs.values():
            if job.status == Status.RUNNING:
                return Status.RUNNING
        service = None
        for service in self.services:
            service.check_job(with_update_report=False)
        if service is not None:
            service.check_job(with_update_report=False)
        has_pending = False
        for job in self.jobs.values():
            job.check_job(with_update_report=False)
            if not job.is_complete():
                has_pending = True
            else:
                if job.status != Status.PASSED:
                    return Status.FAILED
        return Status.PENDING if has_pending else Status.PASSED

    def get_status_dot(self) -> str:
        """
        Get's the status dot for the pipeline.
        """
        if self.get_status() == Status.PASSED:
            return "🟢"
        if self.get_status() == Status.FAILED:
            return "🔴"
        if self.get_status() == Status.SKIPPED:
            return "🔵"
        return "🟡"

    def job(
        self,
        name: str,
        command: str,
        *,
        depends_on: List[str] = None,
        **kwargs,
    ) -> Job:
        """
        Creates a :class:`~jaypore_ci.jci.Job` instance based on the
        pipeline/stage that it is being defined in. See
        :class:`~jaypore_ci.jci.Job` for details on what parameters can be
        passed to the job.
        """
        depends_on = [] if depends_on is None else depends_on
        depends_on = [depends_on] if isinstance(depends_on, str) else depends_on
        assert name not in self.jobs, f"{name} already defined"
        assert name not in self.stages, "Stage name cannot match a job's name"
        kwargs, job_kwargs = dict(self.pipe_kwargs), kwargs
        kwargs.update(self.stage_kwargs if self.stage_kwargs is not None else {})
        kwargs.update(job_kwargs)
        if not kwargs.get("is_service"):
            assert command, f"Command: {command}"
        job = Job(
            name=name if name is not None else " ",
            command=command,
            status=Status.PENDING,
            pipeline=self,
            children=[],
            parents=depends_on,
            **kwargs,
        )
        for parent_name in depends_on:
            assert (
                parent_name in self.jobs
            ), f"Parent job has to be defined before a child. Cannot find {parent_name}"
            parent = self.jobs[parent_name]
            assert parent.stage == job.stage, "Cannot have dependencies across stages"
        self.jobs[name] = job
        if kwargs.get("is_service"):
            self.services.append(job)
        return job

    @classmethod
    def env_matrix(cls, **kwargs):
        """
        Return a cartesian product of all the provided kwargs.
        """
        keys = list(sorted(kwargs.keys()))
        for values in product(*[kwargs[key] for key in keys]):
            yield dict(list(zip(keys, values)))

    def __ensure_duplex__(self):
        for name, job in self.jobs.items():
            for parent_name in job.parents:
                parent = self.jobs[parent_name]
                parent.children = list(sorted(set(parent.children).union(set([name]))))

    def run(self):
        """
        Run the pipeline. This is always called automatically when the context
        of the pipeline declaration finishes and so unless you are doing
        something fancy you don't need to call this manually.
        """
        self.__ensure_duplex__()
        # Run stages one by one
        job = None
        for stage in self.stages:
            # --- Trigger starting jobs
            jobs = {name: job for name, job in self.jobs.items() if job.stage == stage}
            for name in {job.name for job in jobs.values() if not job.parents}:
                jobs[name].trigger()
            # --- monitor and ensure all jobs run
            while not all(job.is_complete() for job in jobs.values()):
                for job in jobs.values():
                    job.check_job(with_update_report=False)
                    if not job.is_complete():
                        # If all dependencies are met: trigger
                        if len(job.parents) == 0 or all(
                            jobs[parent_name].is_complete()
                            and jobs[parent_name].status == Status.PASSED
                            for parent_name in job.parents
                        ):
                            job.trigger()
                        elif any(
                            jobs[parent_name].is_complete()
                            and jobs[parent_name].status != Status.PASSED
                            for parent_name in job.parents
                        ):
                            job.status = Status.SKIPPED
                job.check_job()
                time.sleep(self.poll_interval)
            # --- has this stage passed?
            if not all(
                job.is_complete() and job.status == Status.PASSED
                for job in jobs.values()
            ):
                self.logging().error("Stage failed")
                job.update_report()
                break
        self.logging().error("Pipeline passed")
        if job is not None:
            report = job.update_report()
            self.logging().info("Report:", report=report)

    @contextmanager
    def stage(self, name, **kwargs):
        """
        A stage in a pipeline.

        Any kwargs passed to this stage are supplied to jobs created within
        this stage.
        """
        assert name not in self.jobs, "Stage name cannot match a job's name"
        assert name not in self.stages, "Stage names cannot be re-used"
        self.stages.append(name)
        kwargs["stage"] = name
        self.stage_kwargs = kwargs
        yield  # -------------------------
        self.stage_kwargs = None



"""
An email remote.

This is used to report pipeline status via email.
Multiple updates appear as a single thread.
"""
import os
import time
import smtplib
from html import escape as html_escape

from email.headerregistry import Address
from email.message import EmailMessage
from pathlib import Path
from urllib.parse import urlparse


from jaypore_ci.interfaces import Remote, Repo
from jaypore_ci.logging import logger


class Email(Remote):  # pylint: disable=too-many-instance-attributes
    """
    You can send pipeline status via email using this remote. In order to use it you
    can specify the following environment variables in your secrets:

    .. code-block:: console

        JAYPORE_EMAIL_ADDR=email-account@gmail.com
        JAYPORE_EMAIL_PASSWORD=some-app-password
        JAYPORE_EMAIL_TO=myself@gmail.com,mailing-list@gmail.com
        JAYPORE_EMAIL_FROM=noreply@gmail.com

    If you're using something other than gmail, you can specify
    `JAYPORE_EMAIL_HOST` and `JAYPORE_EMAIL_PORT` as well.

    Once that is done you can supply this remote to your pipeline instead of
    the usual gitea one.

    .. code-block:: python

        from jaypore_ci import jci, remotes, repos

        git = repos.Git.from_env()
        email = remotes.Email.from_env(repo=git)
        with jci.Pipeline(repo=git, remote=email) as p:
            pass
            # Do something

    """

    @classmethod
    def from_env(cls, *, repo: Repo) -> "Email":
        """
        Creates a remote instance from the environment.
        """
        remote = urlparse(repo.remote)
        owner = Path(remote.path).parts[1]
        name = Path(remote.path).parts[2].replace(".git", "")
        return cls(
            host=os.environ.get("JAYPORE_EMAIL_HOST", "smtp.gmail.com"),
            port=int(os.environ.get("JAYPORE_EMAIL_PORT", 465)),
            addr=os.environ["JAYPORE_EMAIL_ADDR"],
            password=os.environ["JAYPORE_EMAIL_PASSWORD"],
            email_to=os.environ["JAYPORE_EMAIL_TO"],
            email_from=os.environ.get(
                "JAYPORE_EMAIL_FROM", os.environ["JAYPORE_EMAIL_ADDR"]
            ),
            subject=f"JCI [{owner}/{name}] [{repo.branch} {repo.sha[:8]}]",
            branch=repo.branch,
            sha=repo.sha,
        )

    def __init__(
        self,
        *,
        host: str,
        port: int,
        addr: str,
        password: str,
        email_to: str,
        email_from: str,
        subject: str,
        publish_interval: int = 30,
        **kwargs,
    ):  # pylint: disable=too-many-arguments
        super().__init__(**kwargs)
        # --- customer
        self.host = host
        self.port = port
        self.addr = addr
        self.password = password
        self.email_to = email_to
        self.email_from = email_from
        self.subject = subject
        self.timeout = 10
        self.publish_interval = publish_interval
        # ---
        self.__smtp__ = None
        self.__last_published_at__ = None
        self.__last_report__ = None

    @property
    def smtp(self):
        if self.__smtp__ is None:
            smtp = smtplib.SMTP_SSL(self.host, self.port)
            smtp.ehlo()
            smtp.login(self.addr, self.password)
            self.__smtp__ = smtp
        return self.__smtp__

    def logging(self):
        """
        Return's a logging instance with information about gitea bound to it.
        """
        return logger.bind(addr=self.addr, host=self.host, port=self.port)

    def publish(self, report: str, status: str) -> None:
        """
        Will publish the report via email.

        :param report: Report to write to remote.
        :param status: One of ["pending", "success", "error", "failure",
            "warning"] This is the dot next to each commit in gitea.
        """
        assert status in ("pending", "success", "error", "failure", "warning")
        if (
            self.__last_published_at__ is not None
            and (time.time() - self.__last_published_at__) < self.publish_interval
        ):
            return
        if self.__last_report__ == report:
            return
        self.__last_report__ = report
        self.__last_published_at__ = time.time()
        # Let's send the email
        msg = EmailMessage()
        msg["Subject"] = self.subject
        msg["From"] = Address("JayporeCI", "JayporeCI", self.email_from)
        msg["To"] = self.email_to
        msg.set_content(report)
        msg.add_alternative(
            f"<html><body><pre>{html_escape(report)}</pre></body></html>",
            subtype="html",
        )
        try:
            self.smtp.send_message(msg)
        except Exception as e:  # pylint: disable=broad-except
            self.logging().exception(e)
        self.__last_published_at__ = time.time()
        self.logging().info(
            "Report published",
            subject=self.subject,
            email_from=self.email_from,
            email_to=self.email_to,
        )



from typing import List

from jaypore_ci.interfaces import Repo


class Mock(Repo):
    def __init__(self, *, files_changed, **kwargs):
        super().__init__(**kwargs)
        self.files_changed = files_changed

    def files_changed(self, target: str) -> List[str]:
        "Returns list of files changed between current sha and target"
        return self.files_changed

    @classmethod
    def from_env(cls, **kwargs) -> "Mock":
        """
        Save whatever is provided to kwargs
        """
        return cls(**kwargs)



"""
A mock executor that actually does not run anything.
"""
import uuid

from jaypore_ci.interfaces import Executor, JobStatus
from jaypore_ci.logging import logger


class Mock(Executor):
    """
    Run jobs via docker.

    This will:
        - Create a separate network for each run
        - Run jobs as part of the network
        - Clean up all jobs when the pipeline exits.
    """

    def __init__(self):
        super().__init__()
        self.pipe_id = None
        self.pipeline = None
        self.__log__ = []
        self.__status__ = {}

    def logging(self):
        """
        Returns a logging instance that has executor specific
        information bound to it.
        """
        return logger.bind(pipe_id=self.pipe_id, network_name=self.get_net())

    def set_pipeline(self, pipeline):
        """
        Set executor's pipeline to the given one.

        This will clean up old networks and create new ones.
        """
        if self.pipe_id is not None:
            self.delete_network()
            self.delete_all_jobs()
        self.pipe_id = id(pipeline)
        self.pipeline = pipeline
        self.create_network()

    def __exit__(self, exc_type, exc_value, traceback):
        self.delete_network()
        self.delete_all_jobs()

    def get_net(self):
        """
        Return a network name based on what the curent pipeline is.
        """
        return f"jaypore_{self.pipe_id}" if self.pipe_id is not None else None

    def create_network(self):
        """
        Will create a docker network.

        If it fails to do so in 3 attempts it will raise an
        exception and fail.
        """
        assert self.pipe_id is not None, "Cannot create network if pipe is not set"
        return self.get_net()

    def delete_all_jobs(self):
        """
        Deletes all jobs associated with the pipeline for this
        executor.

        It will stop any jobs that are still running.
        """
        assert self.pipe_id is not None, "Cannot delete jobs if pipe is not set"
        job = None
        for job in self.pipeline.jobs.values():
            if job.run_id is not None and not job.run_id.startswith("pyrun_"):
                self.logging().info("Stop job:", run_id=job.run_id)
                job.check_job(with_update_report=False)
        if job is not None:
            job.check_job()
        self.logging().info("All jobs stopped")

    def delete_network(self):
        """
        Delete the network for this executor.
        """
        assert self.pipe_id is not None, "Cannot delete network if pipe is not set"
        self.logging().info("Delete network", net=self.get_net())

    def get_job_name(self, job):
        """
        Generates a clean job name slug.
        """
        name = "".join(
            l
            for l in job.name.lower().replace(" ", "_")
            if l in "abcdefghijklmnopqrstuvwxyz_1234567890"
        )
        return name

    def run(self, job: "Job") -> str:
        """
        Run the given job and return a docker container ID.
        """
        assert self.pipe_id is not None, "Cannot run job if pipe id is not set"
        self.pipe_id = id(job.pipeline) if self.pipe_id is None else self.pipe_id
        if not job.is_service:
            assert job.command
        name = self.get_job_name(job)
        if name in self.__status__:
            return None
        run_id = uuid.uuid4().hex
        self.__log__.append((name, run_id, "Run"))
        self.__status__[name] = self.__status__[run_id] = True
        self.logging().info(
            "Run job",
            run_id=run_id,
            env_vars=job.get_env(),
            is_service=job.is_service,
            name=self.get_job_name(job),
            net=self.get_net(),
            image=job.image,
            command=job.command if not job.is_service else None,
        )
        return run_id

    def get_status(self, run_id: str) -> JobStatus:
        """
        Given a run_id, it will get the status for that run.
        """
        status = JobStatus(True, None, "", None, None)
        if run_id in self.__status__:
            status = JobStatus(False, 0, "fake logs", None, None)
        return status

    def get_execution_order(self):
        return {name: i for i, (name, *log) in enumerate(self.__log__)}



"""
A gitea remote git host.

This is used to report pipeline status to the remote.
"""
import os
from pathlib import Path
from urllib.parse import urlparse

import requests

from jaypore_ci.interfaces import Remote, RemoteApiFailed, Repo
from jaypore_ci.logging import logger


class Gitea(Remote):  # pylint: disable=too-many-instance-attributes
    """
    The remote implementation for gitea.
    """

    @classmethod
    def from_env(cls, *, repo: Repo) -> "Gitea":
        """
        Creates a remote instance from the environment.
        It will:

            - Find the remote location using `git remote`.
            - Find the current branch
            - Create a new pull request for that branch
            - Allow posting updates using the gitea token provided
        """
        os.environ["JAYPORE_COMMIT_BRANCH"] = repo.branch
        os.environ["JAYPORE_COMMIT_SHA"] = repo.sha
        remote = urlparse(repo.remote)
        return cls(
            root=f"{remote.scheme}://{remote.netloc}",
            owner=Path(remote.path).parts[1],
            repo=Path(remote.path).parts[2].replace(".git", ""),
            branch=repo.branch,
            token=os.environ["JAYPORE_GITEA_TOKEN"],
            sha=repo.sha,
        )

    def __init__(
        self, *, root, owner, repo, token, **kwargs
    ):  # pylint: disable=too-many-arguments
        super().__init__(**kwargs)
        # --- customer
        self.root = root
        self.api = f"{root}/api/v1"
        self.owner = owner
        self.repo = repo
        self.token = token
        self.timeout = 10
        self.base_branch = "main"
        # ---
        self.__pr_id__ = None

    def logging(self):
        """
        Return's a logging instance with information about gitea bound to it.
        """
        return logger.bind(
            root=self.root, owner=self.owner, repo=self.repo, branch=self.branch
        )

    def get_pr_id(self):
        """
        Returns the pull request ID for the current branch.
        """
        if self.__pr_id__ is None:
            r = requests.post(
                f"{self.api}/repos/{self.owner}/{self.repo}/pulls",
                params={"access_token": self.token},
                timeout=self.timeout,
                json={
                    "base": self.base_branch,
                    "body": "Branch auto created by JayporeCI",
                    "head": self.branch,
                    "title": self.branch,
                },
            )
            self.logging().debug("Get PR Id", status_code=r.status_code)
            if r.status_code == 409:
                self.__pr_id__ = r.text.split("issue_id:")[1].split(",")[0].strip()
                return self.get_pr_id()
            if r.status_code == 201:
                return self.get_pr_id()
            if (
                r.status_code == 404
                and r.json()["message"] == "IsBranchExist"
                and self.base_branch != "develop"
            ):
                self.base_branch = "develop"
                return self.get_pr_id()
            self.logging().debug()(
                "Failed gitea api",
                api=self.api,
                owner=self.owner,
                repo=self.repo,
                token=self.token,
                branch=self.branch,
                status=r.status_code,
                response=r.text,
            )
            raise RemoteApiFailed(r)
        return self.__pr_id__

    def publish(self, report: str, status: str):
        """
        Will publish the report to the remote.

        :param report: Report to write to remote.
        :param status: One of ["pending", "success", "error", "failure",
            "warning"] This is the dot next to each commit in gitea.
        """
        assert status in ("pending", "success", "error", "failure", "warning")
        issue_id = self.get_pr_id()
        # Get existing PR body
        r = requests.get(
            f"{self.api}/repos/{self.owner}/{self.repo}/pulls/{issue_id}",
            timeout=self.timeout,
            params={"access_token": self.token},
        )
        self.logging().debug("Get existing body", status_code=r.status_code)
        assert r.status_code == 200
        body = r.json()["body"]
        body = (line for line in body.split("\n"))
        prefix = []
        for line in body:
            if "```jayporeci" in line:
                prefix = prefix[:-1]
                break
            prefix.append(line)
        while prefix and prefix[-1].strip() == "":
            prefix = prefix[:-1]
        prefix.append("")
        # Post new body with report
        report = "\n".join(prefix) + "\n" + report
        r = requests.patch(
            f"{self.api}/repos/{self.owner}/{self.repo}/pulls/{issue_id}",
            data={"body": report},
            timeout=self.timeout,
            params={"access_token": self.token},
        )
        self.logging().debug("Published new report", status_code=r.status_code)
        # Set commit status
        r = requests.post(
            f"{self.api}/repos/{self.owner}/{self.repo}/statuses/{self.sha}",
            json={
                "context": "JayporeCi",
                "description": f"Pipeline status is: {status}",
                "state": status,
                "target_url": f"{self.root}/{self.owner}/{self.repo}/pulls/{issue_id}",
            },
            timeout=self.timeout,
            params={"access_token": self.token},
        )
        self.logging().debug(
            "Published new status", status=status, status_code=r.status_code
        )



from jaypore_ci.interfaces import Reporter, Status


def __node_mod__(nodes):
    mod = 1
    if len(nodes) > 5:
        mod = 2
    if len(nodes) > 10:
        mod = 3
    return mod


class Markdown(Reporter):
    def __init__(self, *, graph_direction: str = "TD", **kwargs):
        super().__init__(**kwargs)
        self.graph_direction = graph_direction

    def render(self, pipeline):
        """
        Returns a markdown report for a given pipeline.

        It will include a mermaid graph and a collapsible list of logs for each
        job.
        """
        return f"""
<details>
    <summary>JayporeCi: {pipeline.get_status_dot()} {pipeline.remote.sha[:10]}</summary>

{self.__render_graph__(pipeline)}

</details>"""

    def __render_graph__(self, pipeline) -> str:  # pylint: disable=too-many-locals
        """
        Render a mermaid graph given the jobs in the pipeline.
        """
        st_map = {
            Status.PENDING: "pending",
            Status.RUNNING: "running",
            Status.FAILED: "failed",
            Status.PASSED: "passed",
            Status.TIMEOUT: "timeout",
            Status.SKIPPED: "skipped",
        }
        mermaid = f"""
```mermaid
flowchart {self.graph_direction}
"""
        for stage in pipeline.stages:
            nodes, edges = set(), set()
            for job in pipeline.jobs.values():
                if job.stage != stage:
                    continue
                nodes.add(job.name)
                edges |= {(p, job.name) for p in job.parents}
            mermaid += f"""
            subgraph {stage}
                direction {self.graph_direction}
            """
            ref = {n: f"{stage}_{i}" for i, n in enumerate(nodes)}
            # If there are too many nodes, scatter them with different length arrows
            mod = __node_mod__([n for n in nodes if not pipeline.jobs[n].parents])
            for i, n in enumerate(nodes):
                n = pipeline.jobs[n]
                if n.parents:
                    continue
                arrow = "." * ((i % mod) + 1)
                arrow = f"-{arrow}->"
                mermaid += f"""
                s_{stage}(( )) {arrow} {ref[n.name]}({n.name}):::{st_map[n.status]}"""
            mod = __node_mod__([n for n in nodes if pipeline.jobs[n].parents])
            for i, (a, b) in enumerate(edges):
                a, b = pipeline.jobs[a], pipeline.jobs[b]
                arrow = "." * ((i % mod) + 1)
                arrow = f"-{arrow}->"
                mermaid += "\n"
                mermaid += (
                    "                "
                    "{ref[a.name]}({a.name}):::{st_map[a.status]}"
                    "{arrow}"
                    "{ref[b.name]}({b.name}):::{st_map[b.status]}"
                )
            mermaid += """
            end
            """
        for s1, s2 in zip(pipeline.stages, pipeline.stages[1:]):
            mermaid += f"""
            {s1} ---> {s2}
            """
        mermaid += """

            classDef pending fill:#aaa, color:black, stroke:black,stroke-width:2px,stroke-dasharray: 5 5;
            classDef skipped fill:#aaa, color:black, stroke:black,stroke-width:2px;
            classDef assigned fill:#ddd, color:black, stroke:black,stroke-width:2px;
            classDef running fill:#bae1ff,color:black,stroke:black,stroke-width:2px,stroke-dasharray: 5 5;
            classDef passed fill:#88d8b0, color:black, stroke:black;
            classDef failed fill:#ff6f69, color:black, stroke:black;
            classDef timeout fill:#ffda9e, color:black, stroke:black;
``` """
        return mermaid



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jaypore_ci',
 'jaypore_ci.executors',
 'jaypore_ci.remotes',
 'jaypore_ci.reporters',
 'jaypore_ci.repos']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'structlog>=22.3.0,<23.0.0']

setup_kwargs = {
    'name': 'jaypore-ci',
    'version': '0.2.13',
    'description': '',
    'long_description': 'None',
    'author': 'arjoonn sharma',
    'author_email': 'arjoonn.94@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
A mock remote.

This is used to test pipelines.
"""
from jaypore_ci.interfaces import Remote, Repo
from jaypore_ci.logging import logger


class Mock(Remote):  # pylint: disable=too-many-instance-attributes
    """
    A mock remote implementation.
    """

    @classmethod
    def from_env(cls, *, repo: Repo):
        return cls(branch=repo.branch, sha=repo.sha)

    def logging(self):
        """
        Return's a logging instance with information about gitea bound to it.
        """
        return logger.bind(branch=self.branch)

    def get_pr_id(self):
        """
        Returns the pull request ID for the current branch.
        """
        return self.branch

    def publish(self, report: str, status: str):
        """
        Will publish the report to the remote.
        """
        pr_id = self.get_pr_id()
        self.logging().debug(
            "Published report", report=report, status=status, pr_id=pr_id
        )
